/**************************************************************************
Golden Model for GS817
10-31-01 1. Fixed MSB address typo for X36 SDR Mode.
	 2. Added read delay to stop end of cycle read glitch.
12-06-01 1. Changed output pipeline changed to switch PL and FT on the fly.
	 2. Added separate pipeline for coh. data to fix DLW WWWR condition.
	 3. Simplified output control SDR/DDR/FT by adding extra output register.
         4. In timing section, changed tKHQV to tKHQV_FT in flow-thru section.
10-25-02 1. Remove ZZ and CKE for mask change.
         2. Disable Byte Write function in DDR mode.
11-04-02 1. Add extra register to CQ generator to model power up correctly.
11-26-02 1. Removed Byte Write feature from DDR mode
**************************************************************************
 05-05-05 Rev 2.0 1) changed coherency mux logic to detect byte writes correctly.
**************************************************************************/
/*
	Sigma Ram Verilog Bus definitions
	1) The Enable Pins, E1 E2 and E3, have been bussed together.  
           The verilog stimulus bus definition is:  
	   <Enable bus name>[0]->E2, 
           <Enable bus name>[1]->E3
	   <Enable bus name>[2]->E1.  
	2) The EP pins, EP2 and EP3 have been bussed together.  The 
	   verilog stimulus bus definition is: 
	   <Enable Program bus name>[0]->EP2
	   <Enable Program bus name>[1]->EP3
	3) The Mode pins, M2 M3 and M4 have been bussed together.  The 
	   verilog stimulus bus definition is: 
	   <Mode bus name>[0]->M2
	   <Mode bus name>[1]->M3
	   <Mode bus name>[2]->M4
	4) The Byte Write pins, SBA-H have been bussed into two separate (SB)
 	   buses.  The two verilog stimulus bus definition are:
		a. <SB left bus name>[0]->SBD
		   <SB left bus name>[1]->SBH 
		   <SB left bus name>[2]->SBC
		   <SB left bus name[3]->SBG  
		   This bus corresponds to the Byte Writes on the LEFT side of 
		   the package.
		b. <SB right bus name>[0]->SBE
		   <SB right bus name>[1]->SBA 
		   <SB right bus name>[2]->SBF
		   <SB right bus name[3]->SBB.  
		   This bus corresponds to the Byte Writes on the RIGHT side of 
		   the package
	5) The Addresses have been bussed together.  The verilog stimulus 
	   bus definition is: 
	   <Address bus name>[19:0]->A19:A0.  
	   This is not true for DDR mode the bus definition is: 
	   <Address bus name>[19:1]->A19-A1, 
	   <Address bus name>[0]->MCL.
*/
/* 1) Uncomment the speed the model is to run at.  
   2) If DDR is required uncomment the define DDR option.  
   3) If EW or LW Flow through mode is desired then uncomment the FLOW option.
   4) If neither DDR or Flow Though mode is desired then leave the two defines
      commented out.
*/

`timescale 100ps / 100ps
`define DDR
`define	SP333MHZ
//`define	SP300MHZ
//`define	SP275MHZ
//`define	SP250MHZ

// Macros for address topology
`define a01 a[1:0]
`define acp a[2]
`define acq a[3]
`define abp a[5:4]
`define abq a[7:6]
`define ass a[9:8]
`define arp a[12:10]
`define arq a[15:13]
`define abk a[17:16]
`define abnk a[19:16]
`define ablk a[7:4]


// Macros for mode choices.
`define ew_ft 3'b000
`define lw_ft 3'b100
`define ddr   3'b110
`define ew    3'b001
`define dlw   3'b101
`define lw    3'b011

//`define tCHECK 39 //flowthrough mode 250MHZ  Check out mode only
//`define tCHECK 13 //pipeline mode 250MHZ       Check out mode only
`define tCHECK 0 //pipeline mode 250MHZ
module golden (a, adv, ck, e, ep, mcld6, mclt6, m, sb_l, sb_r, sd, tck, tdi, tms, wb, zq, tdo, 
	       cq2b, cq1b, cq2, cq1, DQd, DQh, DQc, DQg, DQe, DQa, DQf, DQb);
   // e(3), ep(2), cq_l(cq2), cq_r(cq1), cqb_l(cq2b), cqb_r(cq1b)
   // lbob(mclt6), zz(mclp6), gb(mcld6), cke(mclk6)
   // dq_r0(DQe), dq_r1(DQa), dq_r2(DQf), dq_r3(DQb), dq_l0(DQd), dq_l1(DQh), dq_l2(DQc), dq_l3(DQg)
   // **** MCLs and MCHs FOR DDR MODE ONLY BELOW ****
   // a[0](MCLW6), sd(MCHN6), sb_l[0](MCLC4), sb_l[2](MCLB3), sb_r[3](MCLB8), sb_r[1](MCLC9)
   // Parameters for states.
   parameter S0 = 4'h0; // Deselect Cycle
   parameter S2 = 4'h2; // Bank Deselect Cycle
   parameter S4 = 4'h4; // Read Cycle, Begin Burst
   parameter S5 = 4'h5; // Read Cycle, Continue Burst
   parameter S6 = 4'h6; // Write Cycle, Begin Burst
   parameter S8 = 4'h8; // Write Cycle, Continue Burst
   parameter SA = 4'hA; // Impossible state (for checking)
   //NOTE: all timing is in 100ps increments
   parameter //Common timing between all modes & Setup/Hold
      `ifdef SP333MHZ
	     tKHKH=30, tKHQV=16, tKHQZ=16, tKHQX=5, tKHQX1=5, tKHCX1=5,
             tKHCH=15, tKLCL=15, tCHQX=-4, tCLQX=-2,tCHQV=4,  tCLQV=2,
             tKHCZ=15, tKHKL=12, tKLKH=12, tAVKH=6, tEVKH=6,  tWVKH=6, tBVKH=6,
	     tadvVKH=6,tKHAX=4,  tKHEX=4,  tKHWX=4, tKHBX=4,  tKHadvX=4;
      `endif
      `ifdef SP300MHZ
	     tKHKH=33, tKHQV=18, tKHQZ=18, tKHQX=5, tKHQX1=5, tKHCX1=5,
	     tKHCH=17, tKLCL=17, tCHQX=-4, tCLQX=-2,tCHQV=4,  tCLQV=2,
             tKHCZ=17, tKHKL=13, tKLKH=13, tAVKH=7, tEVKH=7,  tWVKH=7, tBVKH=7,
	     tadvVKH=7,tKHAX=4,  tKHEX=4,  tKHWX=4, tKHBX=4,  tKHadvX=4;
      `endif
      `ifdef SP275MHZ
	     tKHKH=36, tKHQV=19, tKHQZ=19, tKHQX=5,  tKHQX1=5, tKHCX1=5,
             tKHCH=18, tKLCL=18, tCHQX=-5, tCLQX=-25,tCHQV=5,  tCLQV=25,
             tKHCZ=18, tKHKL=14, tKLKH=14, tAVKH=7,  tEVKH=7,  tWVKH=7, tBVKH=7,
	     tadvVKH=7,tKHAX=5,  tKHEX=5,  tKHWX=5,  tKHBX=5,  tKHadvX=5;
      `endif
      `ifdef SP250MHZ
             tKHKH=40, tKHQV=21, tKHQZ=21, tKHQX=5,  tKHQX1=5, tKHCX1=5,
             tKHCH=20, tKLCL=20, tCHQX=-5, tCLQX=-25,tCHQV=5,  tCLQV=25,
             tKHCZ=20, tKHKL=16, tKLKH=16, tAVKH=8,  tEVKH=8,  tWVKH=8, tBVKH=8,
   	     tadvVKH=8,tKHAX=5,  tKHEX=5,  tKHWX=5,  tKHBX=5,  tKHadvX=5;
      `endif
   parameter // echo clock specific timing & Data Setup/Hold
      `ifdef DDR //echo clock DDR timing & Data Setup/Hold
	 `ifdef SP333MHZ
	    tCHCL=tKHKL+1, tCLCH=tKHKL+1, tDVKH=32, tKHDX=27,
	    tDVKL=32, tKLDX=27;
	 `endif
	 `ifdef SP300MHZ
	    tCHCL=tKHKL+1, tCLCH=tKHKL+1, tDVKH=35, tKHDX=30,
	    tDVKL=35, tKLDX=30;
	 `endif
	 `ifdef SP275MHZ
	    tCHCL=tKHKL+1, tCLCH=tKHKL+1, tDVKH=37, tKHDX=33,
	    tDVKL=37, tKLDX=33;
	 `endif
	 `ifdef SP250MHZ
	    tCHCL=tKHKL+1, tCLCH=tKHKL+1, tDVKH=40, tKHDX=35,
	    tDVKL=40, tKLDX=35;
	 `endif
      `else  //echo clock non-DDR timing & Data Setup/Hold
	 `ifdef SP333MHZ
	    tCHCL=tKHKL+1, tCLCH=tKHKL+1, tDVKH=6, tKHDX=4,
	    tDVKL=6, tKLDX=4;
	 `endif
	 `ifdef SP300MHZ
	    tCHCL=tKHKL+1, tCLCH=tKHKL+1, tDVKH=7, tKHDX=4,
	    tDVKL=7, tKLDX=4;
	 `endif
	 `ifdef SP275MHZ
	    tCHCL=tKHKL+1, tCLCH=tKHKL+1, tDVKH=7, tKHDX=5,
	    tDVKL=7, tKLDX=5;
	 `endif
	 `ifdef SP250MHZ
	    tCHCL=tKHKL+1, tCLCH=tKHKL+1, tDVKH=8, tKHDX=5,
	    tDVKL=8, tKLDX=5;
	 `endif
       `endif 

input adv, ck, mcld6, mclt6, sd, tck, tdi, tms, wb, zq;
input [2:0] m, e;
input [3:0] sb_l, sb_r;
input [1:0] ep;
input [19:0] a;
inout  tdo, cq2, cq2b, cq1, cq1b;
inout  [8:0] DQd, DQh, DQc, DQg;
inout  [8:0] DQe, DQa, DQf, DQb;

tri  tdo, cq2, cq2b, cq1, cq1b;
tri  [8:0] DQd, DQh, DQc, DQg;
tri  [8:0] DQe, DQa, DQf, DQb;

// Signal for various decoded lines.
wire enb, ft, ew, lw, dlw, ddr;
wire x18, x36, x72;
wire read, g, g_pl, g_ft;
wire mat0_rr, mat1_rr, mat2_rr;
wire mat0l_rf, mat0l_fr, mat1l_rf, mat1l_fr;
wire mat0i_rf, mat0i_fr, mat1i_rf, mat1i_fr;
wire mat0_rf, mat0_fr, mat1_rf, mat1_fr;
wire [3:0] mux0rr_l, mux1rr_l, mux2rr_l;
wire [3:0] mux0rr_r, mux1rr_r, mux2rr_r;
wire [3:0] mux0rf_l, mux0fr_l, mux1rf_l, mux1fr_l;
wire [3:0] mux0rf_r, mux0fr_r, mux1rf_r, mux1fr_r;
wire [3:0] nomatr_l, nomatf_l;
wire [3:0] nomatr_r, nomatf_r;
supply0 vss;
supply1 vdd;

wire [8:0] DQa_check, DQb_check, DQc_check, DQd_check, DQe_check, DQf_check, DQg_check, DQh_check;

// Define 18Mb memory array.
reg [8:0] data_l0 [262143:0];
reg [8:0] data_l1 [262143:0];
reg [8:0] data_l2 [262143:0];
reg [8:0] data_l3 [262143:0];
reg [8:0] data_r0 [262143:0];
reg [8:0] data_r1 [262143:0];
reg [8:0] data_r2 [262143:0];
reg [8:0] data_r3 [262143:0];

// Data output and pipeline registers.
wire  [8:0] dor_l0, dor_l1, dor_l2, dor_l3;
wire  [8:0] dor_r0, dor_r1, dor_r2, dor_r3;
reg   [8:0] dor_l0a, dor_l1a, dor_l2a, dor_l3a;
reg   [8:0] dor_r0a, dor_r1a, dor_r2a, dor_r3a;
reg   [8:0] dor_l0b, dor_l1b, dor_l2b, dor_l3b;
reg   [8:0] dor_r0b, dor_r1b, dor_r2b, dor_r3b;
reg   [8:0] do_l0, do_l1, do_l2, do_l3;
reg   [8:0] do_r0, do_r1, do_r2, do_r3;
wire  [8:0] dof_l0, dof_l1, dof_l2, dof_l3;
wire  [8:0] dof_r0, dof_r1, dof_r2, dof_r3;
reg   [8:0] dof_l0a, dof_l1a, dof_l2a, dof_l3a;
reg   [8:0] dof_r0a, dof_r1a, dof_r2a, dof_r3a;
reg   [8:0] dof_l0b, dof_l1b, dof_l2b, dof_l3b;
reg   [8:0] dof_r0b, dof_r1b, dof_r2b, dof_r3b;
reg   [8:0] dof_l0c, dof_l1c, dof_l2c, dof_l3c;
reg   [8:0] dof_r0c, dof_r1c, dof_r2c, dof_r3c;

// Data input and pipeline registers.
wire  [8:0] dir_l0,  dir_l1,  dir_l2,  dir_l3;
wire  [8:0] dir_r0,  dir_r1,  dir_r2,  dir_r3;
reg   [8:0] dir_l0a, dir_l1a, dir_l2a, dir_l3a;
reg   [8:0] dir_r0a, dir_r1a, dir_r2a, dir_r3a;
reg   [8:0] dir_l0b, dir_l1b, dir_l2b, dir_l3b;
reg   [8:0] dir_r0b, dir_r1b, dir_r2b, dir_r3b;
reg   [8:0] dir_l0c, dir_l1c, dir_l2c, dir_l3c;
reg   [8:0] dir_r0c, dir_r1c, dir_r2c, dir_r3c;
reg   [8:0] dcr_l0b, dcr_l1b, dcr_l2b, dcr_l3b;
reg   [8:0] dcr_r0b, dcr_r1b, dcr_r2b, dcr_r3b;
reg   [8:0] dcr_l0c, dcr_l1c, dcr_l2c, dcr_l3c;
reg   [8:0] dcr_r0c, dcr_r1c, dcr_r2c, dcr_r3c;
reg   [8:0] dif_l0a, dif_l1a, dif_l2a, dif_l3a;
reg   [8:0] dif_r0a, dif_r1a, dif_r2a, dif_r3a;
reg   [8:0] dif_l0b, dif_l1b, dif_l2b, dif_l3b;
reg   [8:0] dif_r0b, dif_r1b, dif_r2b, dif_r3b;
reg   [8:0] dcf_l0b, dcf_l1b, dcf_l2b, dcf_l3b;
reg   [8:0] dcf_r0b, dcf_r1b, dcf_r2b, dcf_r3b;

// Address input and pipeline registers.
wire [19:0] wadd;
reg [19:0] radd, wadd_1, wadd_2, wadd_3, wadd_4;

// Byte write input and pipeline registers.
wire [3:0] sbl, sbr;
reg [3:0] sbl_1, sbr_1;
reg [3:0] sbl_2, sbr_2;
reg [3:0] sbl_3, sbr_3;
reg [3:0] sbl_4, sbr_4;

wire mclk6 = 0;
wire mclp6 = 0;
wire en_dqa, en_cqa;
wire en_x72, en_x18, en_x36, kwr, kwf, kwc, write;
reg write_a, write_b, write_c, write_p, en_dqb;
reg en_ft, en_pl, en_cq, en_cqp1, en_cqp2, en_cqp3;
reg [3:0] state, next_state;
reg [1:0] msb;
reg clk_rd, clk_cq, clk_cqa, lsb_r, lsb_w, pok;

// Figure out pipe stages needed from mode input.
assign ew  =  1'b1 ? (m == `ew_ft || m == `ew) : 1'b0;
assign lw  =  1'b1 ? (m == `lw_ft || m == `lw || m == `ddr) : 1'b0;
assign ddr =  1'b1 ? (m == `ddr) : 1'b0;
assign dlw =  1'b1 ? (m == `dlw) : 1'b0;
assign ft  =  1'b1 ? (m == `ew_ft || m == `lw_ft) : 1'b0;

// Define enable for bank select.
assign enb = (e[0] ^ ep[0]) & (e[1] ^ ep[1]);

// Send address to array, also choose which pipe stage
// depending on lw, ew, dlw, etc. Use one less stage than on chip.
assign wadd = ({20{ew}} & wadd_2 | {20{lw}} & wadd_3 | {20{dlw}} & wadd_4);
assign sbl = {4{~ddr}} & ({4{ew}} & sbl_2 | {4{lw}} & sbl_3 | {4{dlw}} & sbl_4);
assign sbr = {4{~ddr}} & ({4{ew}} & sbr_2 | {4{lw}} & sbr_3 | {4{dlw}} & sbr_4);

// Figure out which stage of data pipe line to send to array.
assign dir_l0 = ({9{ew | lw | write_a & dlw}} & dir_l0b | {9{~write_a & dlw}} & dir_l0c);
assign dir_l1 = ({9{ew | lw | write_a & dlw}} & dir_l1b | {9{~write_a & dlw}} & dir_l1c);
assign dir_l2 = ({9{ew | lw | write_a & dlw}} & dir_l2b | {9{~write_a & dlw}} & dir_l2c);
assign dir_l3 = ({9{ew | lw | write_a & dlw}} & dir_l3b | {9{~write_a & dlw}} & dir_l3c);
assign dir_r0 = ({9{ew | lw | write_a & dlw}} & dir_r0b | {9{~write_a & dlw}} & dir_r0c);
assign dir_r1 = ({9{ew | lw | write_a & dlw}} & dir_r1b | {9{~write_a & dlw}} & dir_r1c);
assign dir_r2 = ({9{ew | lw | write_a & dlw}} & dir_r2b | {9{~write_a & dlw}} & dir_r2c);
assign dir_r3 = ({9{ew | lw | write_a & dlw}} & dir_r3b | {9{~write_a & dlw}} & dir_r3c);

//Data input clocks.
assign write = (state == S6) || (state == S8);//write condition
assign kwr = ck & ((ew &  write_p) | (lw & write_a) | (dlw & write_c)); //grab DQ data
assign kwf = ~ck & ddr & write_b; //capture falling edge data from DQ
assign kwc = ~(~(~ck & write_b & dlw) & ~(kwr & ~dlw));//shift coherency data pipe
//Data output clocks.
assign #(tKHQV-`tCHECK) g_pl = ~mcld6; //async output enable for pipe-line timing
assign #(tKHQV-`tCHECK) g_ft = ~mcld6;//async output enable for flow-thru timing
assign g = ft & g_ft | ~ft & g_pl; //async output enable - can use tGLQV?
assign en_cqa = pok & ~ft & (state != S2) & en_cqp3;//sync output enable for cq
assign en_dqa = (state == S4) || (state == S5); //read condition
assign read = clk_rd & en_dqa; //Read clock for data from array.
always @ (ck) clk_cq <= #(tKHQV-`tCHECK) ck; //echo clock
always @ (ck) clk_rd <= #1 ck; //delayed clock to read array
always @ (en_dqa) en_ft <= #(tKHQV-`tCHECK) en_dqa; //sync flow-thru output enable
always @ (en_cqa) en_cq <= #(tKHQV-`tCHECK) en_cqa; //sync pipeline echo clock enable
always @ (en_dqb) en_pl <= #(tKHQV-`tCHECK) en_dqb; //sync pipeline DQ enable

always @ (posedge ck) //Advance output register pipeline.
begin
	if (~ft) begin //pipeline data output path
//	sync pipeline echo clock enable for power up case.
		en_cqp1 <= pok;
		en_cqp2 <= en_cqp1;
		en_cqp3 <= en_cqp2;
//	Shift pipeline one cycle - coherency data needed at end of cycle! 
		dor_l0b <= dor_l0a;
		dor_l1b <= dor_l1a;
		dor_l2b <= dor_l2a;
		dor_l3b <= dor_l3a;
		dor_r0b <= dor_r0a;
		dor_r1b <= dor_r1a;
		dor_r2b <= dor_r2a;
		dor_r3b <= dor_r3a;
		dof_l0b <= dof_l0a;
		dof_l1b <= dof_l1a;
		dof_l2b <= dof_l2a;
		dof_l3b <= dof_l3a;
		dof_r0b <= dof_r0a;
		dof_r1b <= dof_r1a;
		dof_r2b <= dof_r2a;
		dof_r3b <= dof_r3a;
// Delay output data from coherency mux by KHQV.
// Use non-blocking assigment but delay by #1 to avoid race condition.
		#1
		dof_l0c <= dof_l0;
		dof_l1c <= dof_l1;
		dof_l2c <= dof_l2;
		dof_l3c <= dof_l3;
		dof_r0c <= dof_r0;
		dof_r1c <= dof_r1;
		dof_r2c <= dof_r2;
		dof_r3c <= dof_r3;
// Shift final output data register by tKHQV.
		do_l0 <= #((tKHQV-`tCHECK) - 1) dor_l0;
		do_l1 <= #((tKHQV-`tCHECK) - 1) dor_l1;
		do_l2 <= #((tKHQV-`tCHECK) - 1) dor_l2;
		do_l3 <= #((tKHQV-`tCHECK) - 1) dor_l3;
		do_r0 <= #((tKHQV-`tCHECK) - 1) dor_r0;
		do_r1 <= #((tKHQV-`tCHECK) - 1) dor_r1;
		do_r2 <= #((tKHQV-`tCHECK) - 1) dor_r2;
		do_r3 <= #((tKHQV-`tCHECK) - 1) dor_r3;
	end
	if (ft) begin //flow-thru data output path
// Bypass pipeline in flow-thru mode by blocking assignment.
		#2 // delay #2 used because read clock already #1 delay from ck.
		dor_l0b = dor_l0a;
		dor_l1b = dor_l1a;
		dor_l2b = dor_l2a;
		dor_l3b = dor_l3a;
		dor_r0b = dor_r0a;
		dor_r1b = dor_r1a;
		dor_r2b = dor_r2a;
		dor_r3b = dor_r3a;
// Delay output data from coherency mux by tKHQV_FT.
		#((tKHQV-`tCHECK) - 2)
		do_l0 = dor_l0;
		do_l1 = dor_l1;
		do_l2 = dor_l2;
		do_l3 = dor_l3;
		do_r0 = dor_r0;
		do_r1 = dor_r1;
		do_r2 = dor_r2;
		do_r3 = dor_r3;
	end
end

always @ (negedge ck) //Advance DDR output register pipeline.
begin
	write_p <= 1'b0;
	write_a <= write;
	write_c <= write_a;
	if (ddr) begin //extra falling edge data for DDR
//additional pipeline to sync data to falling edge and delay by tKHQV
		do_l0 <= #(tKHQV-`tCHECK) dof_l0c;
		do_l1 <= #(tKHQV-`tCHECK) dof_l1c;
		do_l2 <= #(tKHQV-`tCHECK) dof_l2c;
		do_l3 <= #(tKHQV-`tCHECK) dof_l3c;
		do_r0 <= #(tKHQV-`tCHECK) dof_r0c;
		do_r1 <= #(tKHQV-`tCHECK) dof_r1c;
		do_r2 <= #(tKHQV-`tCHECK) dof_r2c;
		do_r3 <= #(tKHQV-`tCHECK) dof_r3c;
	end
end

always @ (posedge kwc) /* Coherency data input registers.
These are separate from data input registers so this data can be advanced
a half cycle earlier in DLW because because coherency data needs to be
available early before final stage output register clock captures data.*/
begin
	if (~ft) begin //pipeline data path
	// Shift and delay second pipeline stage to match output register timing.
	// Delay #1 added to sync data up to last pipe stage.
		dcr_l0b <= #1 dir_l0a;
		dcr_l1b <= #1 dir_l1a;
		dcr_l2b <= #1 dir_l2a;
		dcr_l3b <= #1 dir_l3a;
		dcr_r0b <= #1 dir_r0a;
		dcr_r1b <= #1 dir_r1a;
		dcr_r2b <= #1 dir_r2a;
		dcr_r3b <= #1 dir_r3a;
		dcf_l0b <= #1 dif_l0a;
		dcf_l1b <= #1 dif_l1a;
		dcf_l2b <= #1 dif_l2a;
		dcf_l3b <= #1 dif_l3a;
		dcf_r0b <= #1 dif_r0a;
		dcf_r1b <= #1 dif_r1a;
		dcf_r2b <= #1 dif_r2a;
		dcf_r3b <= #1 dif_r3a;
	// Shift and delay third pipeline stage to match output register timing.
		dcr_l0c <= #1 dcr_l0b;
		dcr_l1c <= #1 dcr_l1b;
		dcr_l2c <= #1 dcr_l2b;
		dcr_l3c <= #1 dcr_l3b;
		dcr_r0c <= #1 dcr_r0b;
		dcr_r1c <= #1 dcr_r1b;
		dcr_r2c <= #1 dcr_r2b;
		dcr_r3c <= #1 dcr_r3b;
	end
	if (ft) begin //flow-thry data path
	// Shift and delay second pipeline stage to match output register timing.
		dcr_l0b <= #1 dir_l0a;
		dcr_l1b <= #1 dir_l1a;
		dcr_l2b <= #1 dir_l2a;
		dcr_l3b <= #1 dir_l3a;
		dcr_r0b <= #1 dir_r0a;
		dcr_r1b <= #1 dir_r1a;
		dcr_r2b <= #1 dir_r2a;
		dcr_r3b <= #1 dir_r3a;
	end
end

//Coherency signals for level 0, 1, and 2.
assign mat0_rr = (ew | lw | dlw) ? (radd == wadd_1) : 1'b0;
assign mat1_rr = (lw | dlw) ? (radd == wadd_2) : 1'b0;
assign mat2_rr =  dlw ? (radd == wadd_3) : 1'b0;

assign mat0l_rf = ~mclt6 ? (radd == {wadd_1[19:2], wadd_1[1:0] + 2'b01}) : 1'b0;
assign mat0i_rf =  mclt6 ? (radd == {wadd_1[19:1], ~wadd_1[0]}) : 1'b0;
assign mat0_rf  =  ddr & (mat0l_rf | mat0i_rf);
assign mat0l_fr = ~mclt6 ? ({radd[19:2], radd[1:0] + 2'b01 } == wadd_1) : 1'b0;
assign mat0i_fr =  mclt6 ? ({radd[19:1], ~radd[0]} == wadd_1) : 1'b0;
assign mat0_fr  =  ddr & (mat0l_fr | mat0i_fr);

assign mat1l_rf = ~mclt6 ? (radd == {wadd_2[19:2], wadd_2[1:0] + 2'b01}) : 1'b0;
assign mat1i_rf =  mclt6 ? (radd == {wadd_2[19:1], ~wadd_2[0]}) : 1'b0;
assign mat1_rf  =  ddr & (mat1l_rf | mat1i_rf);
assign mat1l_fr = ~mclt6 ? ({radd[19:2], radd[1:0] + 2'b01 } == wadd_2) : 1'b0;
assign mat1i_fr =  mclt6 ? ({radd[19:1], ~radd[0]} == wadd_2) : 1'b0;
assign mat1_fr  =  ddr & (mat1l_fr | mat1i_fr);

//Coherehcy signal with byte write added.  Delay #1 to sync with output pipe.
assign #1 mux0rr_l = ({4{~ddr}} & ~sbl_1 | {4{ddr}}) & {4{  mat0_rr }};
assign #1 mux0rf_l = ({4{~ddr}} & ~sbl_1 | {4{ddr}}) & {4{  mat0_rf }};
assign #1 mux0fr_l = ({4{~ddr}} & ~sbl_1 | {4{ddr}}) & {4{  mat0_fr }};
assign #1 mux1rr_l = ({4{~ddr}} & ~sbl_2 | {4{ddr}}) & (~({4{ mat0_rr}} & ~sbl_1) & ({4{ mat1_rr }} & ~sbl_2));
assign #1 mux1rf_l = ({4{~ddr}} & ~sbl_2 | {4{ddr}}) & (~({4{ mat0_rr}} & ~sbl_1) & ({4{ mat1_rf }} & ~sbl_2));
assign #1 mux1fr_l = ({4{~ddr}} & ~sbl_2 | {4{ddr}}) & (~({4{ mat0_rr}} & ~sbl_1) & ({4{ mat1_fr }} & ~sbl_2));
assign #1 mux2rr_l = ({4{~ddr}} & ~sbl_3 | {4{ddr}}) & (~({4{ mat0_rr}} & ~sbl_1) & ~({4{ mat1_rr }} & ~sbl_2) & ({4{mat2_rr}} & ~sbl_3));
/*
 assign #1 mux0rr_l = ({4{~ddr}} & ~sbl_1 | {4{ddr}}) & {4{  mat0_rr }};
 assign #1 mux0rf_l = ({4{~ddr}} & ~sbl_1 | {4{ddr}}) & {4{  mat0_rf }};
 assign #1 mux0fr_l = ({4{~ddr}} & ~sbl_1 | {4{ddr}}) & {4{  mat0_fr }};
 assign #1 mux1rr_l = ({4{~ddr}} & ~sbl_2 | {4{ddr}}) & {4{ ~mat0_rr & mat1_rr }};
 assign #1 mux1rf_l = ({4{~ddr}} & ~sbl_2 | {4{ddr}}) & {4{ ~mat0_rr & mat1_rf }};
 assign #1 mux1fr_l = ({4{~ddr}} & ~sbl_2 | {4{ddr}}) & {4{ ~mat0_rr & mat1_fr }};
 assign #1 mux2rr_l = ({4{~ddr}} & ~sbl_3 | {4{ddr}}) & {4{ ~mat0_rr & ~mat1_rr & mat2_rr }};
*/
assign nomatr_l = ~mux0rr_l & ~mux1rr_l & ~mux0rf_l & ~mux1rf_l & ~mux2rr_l;
assign nomatf_l = ~mux0rr_l & ~mux1rr_l & ~mux0fr_l & ~mux1fr_l;

assign #1 mux0rr_r = ({4{~ddr}} & ~sbr_1 | {4{ddr}}) & {4{  mat0_rr }};
assign #1 mux0rf_r = ({4{~ddr}} & ~sbr_1 | {4{ddr}}) & {4{  mat0_rf }};
assign #1 mux0fr_r = ({4{~ddr}} & ~sbr_1 | {4{ddr}}) & {4{  mat0_fr }};
assign #1 mux1rr_r = ({4{~ddr}} & ~sbr_2 | {4{ddr}}) & (~({4{ mat0_rr}} & ~sbr_1) & ({4{ mat1_rr }} & ~sbr_2));
assign #1 mux1rf_r = ({4{~ddr}} & ~sbr_2 | {4{ddr}}) & (~({4{ mat0_rr}} & ~sbr_1) & ({4{ mat1_rf }} & ~sbr_2));
assign #1 mux1fr_r = ({4{~ddr}} & ~sbr_2 | {4{ddr}}) & (~({4{ mat0_rr}} & ~sbr_1) & ({4{ mat1_fr }} & ~sbr_2));
assign #1 mux2rr_r = ({4{~ddr}} & ~sbr_3 | {4{ddr}}) & (~({4{ mat0_rr}} & ~sbr_1) & ~({4{ mat1_rr }} & ~sbr_2) & ({4{mat2_rr}} & ~sbr_3));
/*
 assign #1 mux0rr_r = ({4{~ddr}} & ~sbr_1 | {4{ddr}}) & {4{  mat0_rr }};
 assign #1 mux0rf_r = ({4{~ddr}} & ~sbr_1 | {4{ddr}}) & {4{  mat0_rf }};
 assign #1 mux0fr_r = ({4{~ddr}} & ~sbr_1 | {4{ddr}}) & {4{  mat0_fr }};
 assign #1 mux1rr_r = ({4{~ddr}} & ~sbr_2 | {4{ddr}}) & {4{ ~mat0_rr & mat1_rr }};
 assign #1 mux1rf_r = ({4{~ddr}} & ~sbr_2 | {4{ddr}}) & {4{ ~mat0_rr & mat1_rf }};
 assign #1 mux1fr_r = ({4{~ddr}} & ~sbr_2 | {4{ddr}}) & {4{ ~mat0_rr & mat1_fr }};
 assign #1 mux2rr_r = ({4{~ddr}} & ~sbr_3 | {4{ddr}}) & {4{ ~mat0_rr & ~mat1_rr & mat2_rr }};
*/
assign nomatr_r = ~mux0rr_r & ~mux1rr_r & ~mux0rf_r & ~mux1rf_r & ~mux2rr_r;
assign nomatf_r = ~mux0rr_r & ~mux1rr_r & ~mux0fr_r & ~mux1fr_r;

//Final coherency output mux before last output pipe.
assign dor_l0 = dor_l0b & {9{nomatr_l[0]}} | (dir_l0a & {9{mux0rr_l[0]}} | dif_l0a & {9{mux0rf_l[0]}}) |
		({9{~mux0rf_l[0]}} & (dcr_l0b & {9{mux1rr_l[0]}} | dcf_l0b & {9{mux1rf_l[0]}})) | ({9{mux2rr_l[0]}} & dcr_l0c);
assign dor_l1 = dor_l1b & {9{nomatr_l[1]}} | (dir_l1a & {9{mux0rr_l[1]}} | dif_l1a & {9{mux0rf_l[1]}}) |
		({9{~mux0rf_l[1]}} & (dcr_l1b & {9{mux1rr_l[1]}} | dcf_l1b & {9{mux1rf_l[1]}})) | ({9{mux2rr_l[1]}} & dcr_l1c);
assign dor_l2 = dor_l2b & {9{nomatr_l[2]}} | (dir_l2a & {9{mux0rr_l[2]}} | dif_l2a & {9{mux0rf_l[2]}}) |
		({9{~mux0rf_l[2]}} & (dcr_l2b & {9{mux1rr_l[2]}} | dcf_l2b & {9{mux1rf_l[2]}})) | ({9{mux2rr_l[2]}} & dcr_l2c);
assign dor_l3 = dor_l3b & {9{nomatr_l[3]}} | (dir_l3a & {9{mux0rr_l[3]}} | dif_l3a & {9{mux0rf_l[3]}}) |
		({9{~mux0rf_l[3]}} & (dcr_l3b & {9{mux1rr_l[3]}} | dcf_l3b & {9{mux1rf_l[3]}})) | ({9{mux2rr_l[3]}} & dcr_l3c);
assign dor_r0 = dor_r0b & {9{nomatr_r[0]}} | (dir_r0a & {9{mux0rr_r[0]}} | dif_r0a & {9{mux0rf_r[0]}}) |
		({9{~mux0rf_r[0]}} & (dcr_r0b & {9{mux1rr_r[0]}} | dcf_r0b & {9{mux1rf_r[0]}})) | ({9{mux2rr_r[0]}} & dcr_r0c);
assign dor_r1 = dor_r1b & {9{nomatr_r[1]}} | (dir_r1a & {9{mux0rr_r[1]}} | dif_r1a & {9{mux0rf_r[1]}}) |
		({9{~mux0rf_r[1]}} & (dcr_r1b & {9{mux1rr_r[1]}} | dcf_r1b & {9{mux1rf_r[1]}})) | ({9{mux2rr_r[1]}} & dcr_r1c);
assign dor_r2 = dor_r2b & {9{nomatr_r[2]}} | (dir_r2a & {9{mux0rr_r[2]}} | dif_r2a & {9{mux0rf_r[2]}}) |
		({9{~mux0rf_r[2]}} & (dcr_r2b & {9{mux1rr_r[2]}} | dcf_r2b & {9{mux1rf_r[2]}})) | ({9{mux2rr_r[2]}} & dcr_r2c);
assign dor_r3 = dor_r3b & {9{nomatr_r[3]}} | (dir_r3a & {9{mux0rr_r[3]}} | dif_r3a & {9{mux0rf_r[3]}}) |
		({9{~mux0rf_r[3]}} & (dcr_r3b & {9{mux1rr_r[3]}} | dcf_r3b & {9{mux1rf_r[3]}})) | ({9{mux2rr_r[3]}} & dcr_r3c);

assign dof_l0 = dof_l0b & {9{nomatf_l[0]}} | (dif_l0a & {9{mux0rr_l[0]}} | dir_l0a & {9{mux0fr_l[0]}}) |
		({9{~mux0fr_l[0]}} & (dcf_l0b & {9{mux1rr_l[0]}} | dcr_l0b & {9{mux1fr_l[0]}}));
assign dof_l1 = dof_l1b & {9{nomatf_l[1]}} | (dif_l1a & {9{mux0rr_l[1]}} | dir_l1a & {9{mux0fr_l[1]}}) |
		({9{~mux0fr_l[1]}} & (dcf_l1b & {9{mux1rr_l[1]}} | dcr_l1b & {9{mux1fr_l[1]}}));
assign dof_l2 = dof_l2b & {9{nomatf_l[2]}} | (dif_l2a & {9{mux0rr_l[2]}} | dir_l2a & {9{mux0fr_l[2]}}) |
		({9{~mux0fr_l[2]}} & (dcf_l2b & {9{mux1rr_l[2]}} | dcr_l2b & {9{mux1fr_l[2]}}));
assign dof_l3 = dof_l3b & {9{nomatf_l[3]}} | (dif_l3a & {9{mux0rr_l[3]}} | dir_l3a & {9{mux0fr_l[3]}}) |
		({9{~mux0fr_l[3]}} & (dcf_l3b & {9{mux1rr_l[3]}} | dcr_l3b & {9{mux1fr_l[3]}}));
assign dof_r0 = dof_r0b & {9{nomatf_r[0]}} | (dif_r0a & {9{mux0rr_r[0]}} | dir_r0a & {9{mux0fr_r[0]}}) |
		({9{~mux0fr_r[0]}} & (dcf_r0b & {9{mux1rr_r[0]}} | dcr_r0b & {9{mux1fr_r[0]}}));
assign dof_r1 = dof_r1b & {9{nomatf_r[1]}} | (dif_r1a & {9{mux0rr_r[1]}} | dir_r1a & {9{mux0fr_r[1]}}) |
		({9{~mux0fr_r[1]}} & (dcf_r1b & {9{mux1rr_r[1]}} | dcr_r1b & {9{mux1fr_r[1]}}));
assign dof_r2 = dof_r2b & {9{nomatf_r[2]}} | (dif_r2a & {9{mux0rr_r[2]}} | dir_r2a & {9{mux0fr_r[2]}}) |
		({9{~mux0fr_r[2]}} & (dcf_r2b & {9{mux1rr_r[2]}} | dcr_r2b & {9{mux1fr_r[2]}}));
assign dof_r3 = dof_r3b & {9{nomatf_r[3]}} | (dif_r3a & {9{mux0rr_r[3]}} | dir_r3a & {9{mux0fr_r[3]}}) |
		({9{~mux0fr_r[3]}} & (dcf_r3b & {9{mux1rr_r[3]}} | dcr_r3b & {9{mux1fr_r[3]}}));

// DQ tristate mechanism and logic for x18, x36, and x72.
assign en_x18 = x18 & (ft & en_ft | ~ft & en_pl);
assign en_x36 = x36 & (ft & en_ft | ~ft & en_pl);
assign en_x72 = x72 & (ft & en_ft | ~ft & en_pl);

assign DQd = g & (en_x72 | en_x36)          ? do_l0 : 9'bz;
assign DQh = g &  en_x72                    ? do_l1 : 9'bz;
assign DQc = g & (en_x72 | en_x36 | en_x18) ? do_l2 : 9'bz;
assign DQg = g &  en_x72                    ? do_l3 : 9'bz;
assign DQe = g &  en_x72                    ? do_r0 : 9'bz;
assign DQa = g & (en_x72 | en_x36 | en_x18) ? do_r1 : 9'bz;
assign DQf = g &  en_x72                    ? do_r2 : 9'bz;
assign DQb = g & (en_x72 | en_x36)          ? do_r3 : 9'bz;

assign DQd_check = ~(en_x72 | en_x36)          ? DQd : 9'bz;
assign DQh_check = ~en_x72                     ? DQh : 9'bz;
assign DQc_check = ~(en_x72 | en_x36 | en_x18) ? DQc : 9'bz;
assign DQg_check = ~en_x72                     ? DQg : 9'bz;
assign DQe_check = ~en_x72                     ? DQe : 9'bz;
assign DQa_check = ~(en_x72 | en_x36 | en_x18) ? DQa : 9'bz;
assign DQf_check = ~en_x72                     ? DQf : 9'bz;
assign DQb_check = ~(en_x72 | en_x36)          ? DQb : 9'bz;

// CQ for ddr, lw, and ew - no ft, else tristate bus.
assign  cq2  = en_cq ?  clk_cq  : 1'bz;
assign  cq2b = en_cq ? ~clk_cq  : 1'bz;
assign  cq1  = en_cq ?  clk_cq  : 1'bz;
assign  cq1b = en_cq ? ~clk_cq  : 1'bz;

//Option devices from main simulation - enables X18,X36, or X72.
optx72 o72a (x72, vdd);
optx36 o36a (x72, vss);
optx18 o18a (x72, vss);
optx72 o72b (x36, vss);
optx36 o36b (x36, vdd);
optx18 o18b (x36, vss);
optx72 o72c (x18, vss);
optx36 o36c (x18, vss);
optx18 o18c (x18, vdd);

initial //Initial conditions for assorted registers.
begin
	state = 4'b0;
	pok = 1'b0;
	next_state = 4'b0;
	en_dqb = 1'b0;
	clk_cqa = 1'b0;	
	en_cqp1 = 1'b0;
	en_cqp2 = 1'b0;
	en_cqp3 = 1'b0;
	clk_cq = 1'b0;
	clk_rd = 1'b0;
	en_ft = 1'b0;
	en_pl = 1'b0;
	en_cq = 1'b0;
	radd = 20'b0;
	wadd_1 = 20'b0;
	wadd_2 = 20'b0;
	wadd_3 = 20'b0;
	wadd_4 = 20'b0;
	dor_l0a = 9'b0;
	dor_l0b = 9'b0;
	do_l0 = 9'b0;
	dor_l1a = 9'b0;
	dor_l1b = 9'b0;
	do_l1 = 9'b0;
	dor_l2a = 9'b0;
	dor_l2b = 9'b0;
	do_l2 = 9'b0;
	dor_l3a = 9'b0;
	dor_l3b = 9'b0;
	do_l3 = 9'b0;
	dor_r0a = 9'b0;
	dor_r0b = 9'b0;
	do_r0 = 9'b0;
	dor_r1a = 9'b0;
	dor_r1b = 9'b0;
	do_r1 = 9'b0;
	dor_r2a = 9'b0;
	dor_r2b = 9'b0;
	do_r2 = 9'b0;
	dor_r3a = 9'b0;
	dor_r3b = 9'b0;
	do_r3 = 9'b0;
	pok <= #(0.1) 1'b1;
end

// Combinational next state logic for state machine.
always @ (state or adv or e or wb or enb)
begin
	case (state)
	S0: begin //deselect
		if (adv & ~e[2] & ~enb)
			next_state = S0; //deselect
		else if (e[2])
			next_state = S0; //deselect
		else if (enb)
			next_state = S2; //bank deselect
		else if (~adv & wb & ~e[2] & ~enb)
			next_state = S4; //read
		else if(~adv & ~wb & ~e[2] & ~enb)
			next_state = S6; //write
		else
		begin
			next_state = SA; //impossible state
			$display("Invalid next state found at %d from S0!\n", $time);
		end
	end
	S2: begin //bank deselect
		if (e[2])
			next_state = S0; //deselect
		else if (adv & ~e[2] & ~enb)
			next_state <= S2; //deselect
		else if (enb)
			next_state = S2; //bank deselect
		else if (~adv & wb & ~e[2] & ~enb)
			next_state = S4; //read
		else if (~adv & ~wb & ~e[2] & ~enb)
			next_state = S6; //write
		else
		begin
			next_state = SA; //impossible state
			$display("Invalid next state found at %d from S2!\n", $time);
		end
	end
	S4: begin //read
		if (e[2])
			next_state = S0; //deselect
		else if (enb)
			next_state = S2; //bank deselect
		else if (~adv & wb & ~e[2] & ~enb)
			next_state = S4; //read
		else if (adv & ~e[2] & ~enb)
			next_state = S5; //read burst
		else if(~adv & ~wb & ~e[2] & ~enb)
			next_state = S6; //write
		else
		begin
			next_state = SA; //impossible state
			$display("Invalid next state found at %d! from S4!\n", $time);
		end
	end
	S5: begin //read burst
		if (e[2])
			next_state = S0; //deselect
		else if (enb)
			next_state = S2; //bank deselect
		else if (~adv & wb & ~e[2] & ~enb)
			next_state = S4; //read
		else if (adv & ~e[2] & ~enb)
			next_state = S5; //read burst
		else if (~adv & ~wb & ~e[2] & ~enb)
			next_state = S6; //write
		else
		begin
			next_state = SA; //impossible state
			$display("Invalid next state found at %d from S5!\n", $time);
		end
	end
	S6: begin //write
		if (e[2])
			next_state = S0; //deselect
		else if (enb)
			next_state = S2; //bank deselect
		else if (~adv & wb & ~e[2] & ~enb)
			next_state = S4; //read
		else if (~adv & ~wb & ~e[2] & ~enb)
			next_state = S6; //write
		else if (adv)
			next_state = S8; //write burst
		else
		begin
			next_state = SA; //impossible state
			$display("Invalid next state found at %d from S6!\n", $time);
		end
	end
	S8: begin //write burst
		if (e[2])
			next_state = S0; //deselect
		else if (enb)
			next_state = S2; //bank deselect
		else if (~adv & wb & ~e[2] & ~enb)
			next_state = S4; //read
		else if (~adv & ~wb & ~e[2] & ~enb)
			next_state = S6; //write
		else if (adv & ~e[2] & ~enb)
			next_state = S8; //write burst
		else
		begin
			next_state = SA; //impossible state
			$display("Invalid next state found at %d from S8!\n", $time);
		end
	end
	default : $display("Invalid present state at %d!\n", $time);
	endcase
end

// Hardware pipeline for shifting address for ew, lw, & dlw.
always @ (posedge ck)
begin
write_b <= write;
state <= next_state;
en_dqb <= en_dqa;
	if (next_state == S4) //Read:  grab address.
	begin
		radd <=  {20{x18}} & a[19:0] | {20{x36}} & { 1'b0, a[18:0] } | {20{x72}} & { 2'b00, a[17:0] };
		lsb_r <= a[0];
	end
	if (next_state == S5) //Increment read address.
	begin
		if ( ~ddr ) begin
			if ( ~lsb_r & mclt6 )
				radd <= {radd[19:2], radd[1:0] + 2'b01};
			else if ( lsb_r & mclt6)
				radd <= {radd[19:2], radd[1:0] - 2'b01};
			else if ( ~mclt6 )
				radd <= {radd[19:2], radd[1:0] + 2'b01};
			else 
				$display("Problem found with counter during read at %d\n", $time);
		end
		else if ( ddr )
			radd <= {radd[19:2], ~radd[1], radd[0]};
		else 
			$display("Problem found with counter during read at %d\n", $time);
	end
	if (next_state == S6) //Write:  grab address.
	begin
		write_p <= 1'b1;
		lsb_w <= a[0];
		wadd_1 <=  {20{x18}} & a[19:0] | {20{x36}} & { 1'b0, a[18:0] } | {20{x72}} & { 2'b00, a[17:0] };
		wadd_2 <= wadd_1;
		wadd_3 <= wadd_2;
		wadd_4 <= wadd_3;
	end
	if (next_state == S8) //Write:  burst.
	begin
		write_p <= 1'b1;
		if ( ~ddr ) begin
			if ( ~lsb_w & mclt6 )
				wadd_1 <= {wadd_1[19:2], wadd_1[1:0] + 2'b01};
			else if ( lsb_w & mclt6)
				wadd_1 <= {wadd_1[19:2], wadd_1[1:0] - 2'b01};
			else if ( ~mclt6 )
				wadd_1 <= {wadd_1[19:2], wadd_1[1:0] + 2'b01};
			else 
				$display("Problem found with counter during write at %d\n", $time);
		end
		else if ( ddr )
			wadd_1 <= {wadd_1[19:2], ~wadd_1[1], wadd_1[0]};
		else 
			$display("Problem found with counter during write at %d\n", $time);
		wadd_2 <= wadd_1;
		wadd_3 <= wadd_2;
		wadd_4 <= wadd_3;
	end
	if (next_state == S6 || next_state == S8) // Write or write burst:  Grab byte write and write to array.
	begin
		sbl_1 <= sb_l;
		sbr_1 <= sb_r;
		// First stage pileline for EW.
		sbl_2 <= sbl_1;
		sbr_2 <= sbr_1;
		// First stage pileline for LW & DDR.
		sbl_3 <= sbl_2;
		sbr_3 <= sbr_2;
		// Second stage pileline for DLW.
		sbl_4 <= sbl_3;
		sbr_4 <= sbr_3;
		if ( x18 & ~ddr ) begin // Put data into the array (X18 case).
			#1 msb = { wadd[19], wadd[18] };
			case ( msb )
				2'b00 : begin
					if (~sbl[2])
						data_l0[wadd[17:0]] <= dir_l2;
					if (~sbr[1])
						data_r0[wadd[17:0]] <= dir_r1;
				end
				2'b01 : begin
					if (~sbl[2])
						data_l1[wadd[17:0]] <= dir_l2;
					if (~sbr[1])
						data_r1[wadd[17:0]] <= dir_r1;
				end
				2'b10 : begin
					if (~sbl[2])
						data_l2[wadd[17:0]] <= dir_l2;
					if (~sbr[1])
						data_r2[wadd[17:0]] <= dir_r1;
				end
				2'b11 : begin
					if (~sbl[2])
						data_l3[wadd[17:0]] <= dir_l2;
					if (~sbr[1])
						data_r3[wadd[17:0]] <= dir_r1;
				end
			endcase
		end
		else if ( x18 & ddr ) begin // Put data into the array (X18 ddr case).
			#1 msb = { wadd[19], wadd[0] };
			case ( msb )
				2'b00 : begin
					if (~sbl[2]) begin
						data_l0[wadd[18:1]] <= dir_l2;
						data_l1[wadd[18:1]] <= dif_l2b;
					end
					if (~sbr[0]) begin
						data_r0[wadd[18:1]] <= dir_r1;
						data_r1[wadd[18:1]] <= dif_r1b;
					end
				end
				2'b01 : begin
					if (~sbl[2]) begin
						data_l0[{ wadd[18:2], ~mclt6 ^ wadd[1] }] <= dif_l2b;
						data_l1[wadd[18:1]] <= dir_l2;
					end
					if (~sbr[0]) begin
						data_r0[{ wadd[18:2], ~mclt6 ^ wadd[1] }] <= dif_r1b;
						data_r1[wadd[18:1]] <= dir_r1;
					end
				end
				2'b10 : begin
					if (~sbl[2]) begin
						data_l2[wadd[18:1]] <= dir_l2;
						data_l3[wadd[18:1]] <= dif_l2b;
					end
					if (~sbr[0]) begin
						data_r2[wadd[18:1]] <= dir_r1;
						data_r3[wadd[18:1]] <= dif_r1b;
					end
				end
				2'b11 : begin
					if (~sbl[2]) begin
						data_l2[{ wadd[18:2], ~mclt6 ^ wadd[1] }] <= dif_l2b;
						data_l3[wadd[18:1]] <= dir_l2;
					end
					if (~sbr[0]) begin
						data_r2[{ wadd[18:2], ~mclt6 ^ wadd[1] }] <= dif_r1b;
						data_r3[wadd[18:1]] <= dir_r1;
					end
				end
			endcase
		end
		else if ( x36 & ~ddr ) begin // Put data into the array (X36 case).
			#1 msb = { wadd[18], 1'b0 }; //was { wadd[19], 1'b0 } (10-31-01)
			case ( msb[1] )
				1'b0 : begin
					if (~sbl[0])
						data_l0[wadd[17:0]] <= dir_l0;
					if (~sbl[2])
						data_l2[wadd[17:0]] <= dir_l2;
					if (~sbr[1])
						data_r0[wadd[17:0]] <= dir_r1;
					if (~sbr[3])
						data_r2[wadd[17:0]] <= dir_r3;
				end
				1'b1 : begin
					if (~sbl[0])
						data_l1[wadd[17:0]] <= dir_l0;
					if (~sbl[2])
						data_l3[wadd[17:0]] <= dir_l2;
					if (~sbr[1])
						data_r1[wadd[17:0]] <= dir_r1;
					if (~sbr[3])
						data_r3[wadd[17:0]] <= dir_r3;
				end
			endcase
		end
		else if ( x36 & ddr ) begin // Put data into the array (X36 ddr case).
			#1 msb = { 1'b0, wadd[0] };
			case ( msb[0] )
				1'b0 : begin
					if (~sbl[0]) begin
						data_l0[wadd[18:1]] <= dir_l0;
						data_l1[wadd[18:1]] <= dif_l0b;
					end
					if (~sbl[2]) begin
						data_l2[wadd[18:1]] <= dir_l2;
						data_l3[wadd[18:1]] <= dif_l2b;
					end
					if (~sbr[1]) begin
						data_r0[wadd[18:1]] <= dir_r1;
						data_r1[wadd[18:1]] <= dif_r1b;
					end
					if (~sbr[3]) begin
						data_r2[wadd[18:1]] <= dir_r3;
						data_r3[wadd[18:1]] <= dif_r3b;
					end
				end
				1'b1 : begin
					if (~sbl[0]) begin
						data_l1[wadd[18:1]] <= dir_l0;
						data_l0[{wadd[18:2], ~mclt6 ^ wadd[1] }] <= dif_l0b;
					end
					if (~sbl[2]) begin
						data_l3[wadd[18:1]] <= dir_l2;
						data_l2[{wadd[18:2], ~mclt6 ^ wadd[1] }] <= dif_l2b;
					end
					if (~sbr[1]) begin
						data_r1[wadd[18:1]] <= dir_r1;
						data_r0[{wadd[18:2], ~mclt6 ^ wadd[1] }] <= dif_r1b;
					end
					if (~sbr[3]) begin
						data_r3[wadd[18:1]] <= dir_r3;
						data_r2[{wadd[18:2], ~mclt6 ^ wadd[1] }] <= dif_r3b;
					end
				end
			endcase
		end
		else if ( x72 ) begin // Put data into the array (X72 case).
				#1;
				if (~sbl[0])
					data_l0[wadd[17:0]] <= dir_l0;
				if (~sbl[1])
					data_l1[wadd[17:0]] <= dir_l1;
				if (~sbl[2])
					data_l2[wadd[17:0]] <= dir_l2;
				if (~sbl[3])
					data_l3[wadd[17:0]] <= dir_l3;
				if (~sbr[0])
					data_r0[wadd[17:0]] <= dir_r0;
				if (~sbr[1])
					data_r1[wadd[17:0]] <= dir_r1;
				if (~sbr[2])
					data_r2[wadd[17:0]] <= dir_r2;
				if (~sbr[3])
					data_r3[wadd[17:0]] <= dir_r3;
		end
	end
end

always @ (posedge kwr) //Data input register pipeline.
begin
	// First register to grab data.
	dir_l0a <= DQd;
	dir_l1a <= DQh;
	dir_l2a <= DQc;
	dir_l3a <= DQg;
	dir_r0a <= DQe;
	dir_r1a <= DQa;
	dir_r2a <= DQf;
	dir_r3a <= DQb;
	// First stage array pipeline for LW and DDR.
	dir_l0b <= dir_l0a;
	dir_l1b <= dir_l1a;
	dir_l2b <= dir_l2a;
	dir_l3b <= dir_l3a;
	dir_r0b <= dir_r0a;
	dir_r1b <= dir_r1a;
	dir_r2b <= dir_r2a;
	dir_r3b <= dir_r3a;
	dif_l0b <= dif_l0a;
	dif_l1b <= dif_l1a;
	dif_l2b <= dif_l2a;
	dif_l3b <= dif_l3a;
	dif_r0b <= dif_r0a;
	dif_r1b <= dif_r1a;
	dif_r2b <= dif_r2a;
	dif_r3b <= dif_r3a;
	// Second stage array pipeline for DLW.
	dir_l0c <= dir_l0b;
	dir_l1c <= dir_l1b;
	dir_l2c <= dir_l2b;
	dir_l3c <= dir_l3b;
	dir_r0c <= dir_r0b;
	dir_r1c <= dir_r1b;
	dir_r2c <= dir_r2b;
	dir_r3c <= dir_r3b;
end

always @ (posedge kwf) //Capture falling edge data for DDR.
begin
	// First register to grab data.
	dif_l0a <= DQd;
	dif_l1a <= DQh;
	dif_l2a <= DQc;
	dif_l3a <= DQg;
	dif_r0a <= DQe;
	dif_r1a <= DQa;
	dif_r2a <= DQf;
	dif_r3a <= DQb;
end

always @ (posedge read) //Read data from array.
begin
	if ( x18 & ~ddr ) begin
		msb = { radd[19], radd[18] };
		case ( msb )
			2'd0 : begin
				dor_l2a <= data_l0[radd[17:0]];
				dor_r1a <= data_r0[radd[17:0]];
			end
			2'd1 : begin
				dor_l2a <= data_l1[radd[17:0]];
				dor_r1a <= data_r1[radd[17:0]];
			end
			2'd2 : begin
				dor_l2a <= data_l2[radd[17:0]];
				dor_r1a <= data_r2[radd[17:0]];
			end
			2'd3 : begin
				dor_l2a <= data_l3[radd[17:0]];
				dor_r1a <= data_r3[radd[17:0]];
			end
		endcase
	end
	else if ( x18 & ddr ) begin
		msb = { radd[19], radd[0] };
		case ( msb )
			2'd0 : begin
				dor_l2a <= data_l0[radd[18:1]];
				dor_r1a <= data_r0[radd[18:1]];
				dof_l2a <= data_l1[radd[18:1]];
				dof_r1a <= data_r1[radd[18:1]];
			end
			2'd1 : begin
				dor_l2a <= data_l1[radd[18:1]];
				dor_r1a <= data_r1[radd[18:1]];
				dof_l2a <= data_l0[{radd[18:2], ~mclt6 ^ radd[1] }];
				dof_r1a <= data_r0[{radd[18:2], ~mclt6 ^ radd[1] }];
			end
			2'd2 : begin
				dor_l2a <= data_l2[radd[18:1]];
				dor_r1a <= data_r2[radd[18:1]];
				dof_l2a <= data_l3[radd[18:1]];
				dof_r1a <= data_r3[radd[18:1]];
			end
			2'd3 : begin
				dor_l2a <= data_l3[radd[18:1]];
				dor_r1a <= data_r3[radd[18:1]];
				dof_l2a <= data_l2[{radd[18:2], ~mclt6 ^ radd[1] }];
				dof_r1a <= data_r2[{radd[18:2], ~mclt6 ^ radd[1] }];
			end
		endcase
	end
	else if ( x36 & ~ddr ) begin
		msb = { radd[18], 1'b0 }; //was { radd[19], 1'b0 } (10-31-01)
		case ( msb[1] )
			1'b0 : begin
				dor_l0a <= data_l0[radd[17:0]];
				dor_l2a <= data_l2[radd[17:0]];
				dor_r1a <= data_r0[radd[17:0]];
				dor_r3a <= data_r2[radd[17:0]];
			end
			1'b1 : begin
				dor_l0a <= data_l1[radd[17:0]];
				dor_l2a <= data_l3[radd[17:0]];
				dor_r1a <= data_r1[radd[17:0]];
				dor_r3a <= data_r3[radd[17:0]];
			end
		endcase
	end
	else if ( x36 & ddr ) begin
		msb = { 1'b0, radd[0] };
		case ( msb[0] )
			1'b0 : begin
				dor_l0a <= data_l0[radd[18:1]];
				dor_l2a <= data_l2[radd[18:1]];
				dor_r1a <= data_r0[radd[18:1]];
				dor_r3a <= data_r2[radd[18:1]];
				dof_l0a <= data_l1[radd[18:1]];
				dof_l2a <= data_l3[radd[18:1]];
				dof_r1a <= data_r1[radd[18:1]];
				dof_r3a <= data_r3[radd[18:1]];
			end
			1'b1 : begin
				dor_l0a <= data_l1[radd[18:1]];
				dor_l2a <= data_l3[radd[18:1]];
				dor_r1a <= data_r1[radd[18:1]];
				dor_r3a <= data_r3[radd[18:1]];
				dof_l0a <= data_l0[{radd[18:2], ~mclt6 ^ radd[1] }];
				dof_l2a <= data_l2[{radd[18:2], ~mclt6 ^ radd[1] }];
				dof_r1a <= data_r0[{radd[18:2], ~mclt6 ^ radd[1] }];
				dof_r3a <= data_r2[{radd[18:2], ~mclt6 ^ radd[1] }];
			end
		endcase
	end
	else if ( x72 ) begin
		dor_l0a <= data_l0[radd[17:0]];
		dor_l1a <= data_l1[radd[17:0]];
		dor_l2a <= data_l2[radd[17:0]];
		dor_l3a <= data_l3[radd[17:0]];
		dor_r0a <= data_r0[radd[17:0]];
		dor_r1a <= data_r1[radd[17:0]];
		dor_r2a <= data_r2[radd[17:0]];
		dor_r3a <= data_r3[radd[17:0]];
	end
end
endmodule //golden.v

